'' System Version 2.001
'' uALFAT_Interface Changes:
''
''   2006/09/24 - Initial version with change dates

'' This object provides an I2C interface to the GHI Electronics
'' uALFAT-SD MMC/SD card interface processor.

CON
  uALFAT_SCL    = 0                                 ' uALFAT I2C SCL
  uALFAT_SDA    = 1                                 ' uALFAT I2C SDA
  uALFAT_Bsy    = 2                                 ' uALFAT Output Busy (high true)
  uALFAT_Rdy    = 3                                 ' uALFAT Input Ready (high true)
  uALFAT_Rst    = 4                                 ' uALFAT Reset (low true)
  uALFAT_SSel   = 5                                 ' uALFAT SPI Select (high true)
  uALFAT_SClk   = 6                                 ' uALFAT SPI Clock (low true)

  uALFAT_EndDir = $4D                               ' End of directory information
  uALFAT_NDT    = $FF                               ' I2C No Data
  uALFAT_HDT    = $FE                               ' I2C Half Data

  uALFAT_Addr   = %10100100                         ' I2C Device Address

OBJ
  dsp : "OS_dspText"                                ' OS display driver interface
  ldr : "OS_loader"                                 ' OS I2C driver interface

PUB start                                           '' Set up the I/O pins
  dira[uALFAT_Rdy] ~                                ' Set some pins to inputs
  dira[uALFAT_Bsy] ~
  dira[uALFAT_SCL] ~
  dira[uALFAT_SDA] ~
  outa[uALFAT_SSel] ~~                              ' SPI Select to HIGH
  dira[uALFAT_SSel] ~~
  outa[uALFAT_SClk] ~                               ' SPI Clock to LOW
  dira[uALFAT_SClk] ~~
  outa[uALFAT_Rst] ~                                ' Take reset LOW then HIGH
  dira[uALFAT_Rst] ~~
  dira[uALFAT_Rst] ~
  
PUB stop                                            '' Make sure the pins are left as inputs

PUB connectCard(ms) | e                             '' Prepare to access an SD/MMC card
  sendStr(string("I",13))                           ' Attach to the media
  checkErrorCode(ms,0)

PUB listDirectory(ms) | e, d, a, s, n0, n1, n2, n3  '' Display the current directory
  sendStr(string("@",13))                           ' Initialize the directory listing
  checkErrorCode(ms,0)
  dsp.str(string("File Name    Attr     File Size",dsp#Cr))
  repeat
    sendStr(string("N",13))                         ' Get the next file entry
    if checkErrorCode(ms,uALFAT_EndDir)
      quit
    getFileName(ms,@n0)
    checkNext(ms," ")
    a := getHex(ms,2)                               ' Get file attribute
    checkNext(ms," ")
    s := getHex(ms,8)                               ' Get file size
    checkNext(ms,13)
    checkErrorCode(ms,0)
    dsp.str(@n0)                                    ' Display the information
    repeat 13 - strsize(@n0)
      dsp.out(" ")
    dsp.bin(a,8)
    dsp.out(" ")
    dsp.dec(s)
    dsp.str(string(" Bytes",dsp#Cr))
  sendStr(string("K",13))                           ' Get media statistics
  checkErrorCode(ms,0)
  a := getHexLong(ms)                               ' Get media size
  checkNext(ms," ")
  s := getHexLong(ms)                               ' Get free size
  checkNext(ms,13)
  checkErrorCode(ms,0)
  dsp.dec(a)
  dsp.str(string(" Total Size, "))
  dsp.dec(s)
  dsp.str(string(" Bytes Free",dsp#Cr))

PUB changeDir(ms, s) : e                            '' Change to specified subdirectory
  sendStr(string("A "))
  sendStr(s)
  sendChar(13)
  checkErrorCode(ms,0)

PUB openFile(ms, s, mode) : e                       '' Open specified file in the specified mode
  sendStr(string("OF 0"))                           ' Send command and use handle zero
  sendChar(mode)                                    ' Send file mode ("R", "W", "A")
  sendChar(">")
  sendStr(s)                                        ' Send file name
  sendChar(13)
  checkErrorCode(ms,0)

PUB closeFile(ms) | e                               '' Close current file
  sendStr(string("C 0",13))                         ' Send command and use handle zero
  checkErrorCode(ms,0)
  
PUB flushFile(ms) | e                               '' Flush current file
  sendStr(string("F 0",13))                         ' Send command and use handle zero
  checkErrorCode(ms,0)
  
PUB readFile(ms, bufAdr, count) | e, t
  sendStr(string("R 0",$7F,">"))
  sendHex(count,8)
  sendChar(13)
  checkErrorCode(ms,0)
  repeat count
    if (t := recvChar(ms)) == -1
      abort t
    byte[bufAdr++] := t
  if getHexLong(ms) <> count
    abort -1
  checkNext(ms,13)
  checkErrorCode(ms,0)

PUB writeFile(ms, bufAdr, count) | e, t
  sendStr(string("W 0>"))
  sendHex(count,8)
  sendChar(13)
  checkErrorCode(ms,0)
  repeat count
    sendChar(byte[bufAdr++])
  if getHexLong(ms) <> count
    abort -1
  checkNext(ms,13)
  checkErrorCode(ms,0)

PUB deleteFile(ms, nameBuf) | e
  sendStr(string("D "))
  sendStr(nameBuf)
  sendChar(13)
  checkErrorCode(ms,0)

PUB getFileName(ms, nameBuf) | i, name, name2, ext  '' Receive a fixed width 8.3 file name
  getTextField(ms,@name,8)                          ''  suppress trailing blanks and make
  checkNext(ms,".")                                 ''  into a single zero-terminated string
  getTextField(ms,@ext,3)
  i := 0
  repeat while i < 8 and name[i] <> " "
    byte[nameBuf++] := name[i]
  byte[nameBuf++] := "."
  i := 0
  repeat while i < 3 and ext[i] <> " "
    byte[nameBuf++] := ext[i]
  byte[nameBuf] := 0

PUB getHexLong(ms)                                  '' Receive a hex long prefixed by "$"
  checkNext(ms,"$")
  return getHex(ms,8)

PUB getHexWord(ms)                                  '' Receive a hex word prefixed by "$"
  checkNext(ms,"$")
  return getHex(ms,4)

PUB getHexByte(ms)                                  '' Receive a hex byte prefixed by "$"
  checkNext(ms,"$")
  return getHex(ms,2)

PUB getTextField(ms, bufAddr, size) | t             '' Receive a fixed size text field
  repeat size
    if (t := recvChar(ms)) == -1
      abort t
    byte[bufAddr++] := t

PUB checkErrorCode(ms, code)                        '' Make sure error code is zero
  if (result := getErrorCode(ms)) <> 0              ' Allow optional specific code
    if result <> code
      abort $100 | result

PUB getErrorCode(ms) : err                          '' Receive an error code ("!00",13)
  checkNext(ms,"!")
  err := getHex(ms,2)
  checkNext(ms,13)

PUB checkNext(ms,char)                              '' Check the next received character
  if (result := recvChar(ms)) == -1 or result <> char
    abort result

PUB getHex(ms,size) : hex | t                       '' Receive a fixed width hex value
  repeat size
    case t := recvChar(ms)
      "0".."9": hex := (hex << 4) | (t - "0")
      "A".."F": hex := (hex << 4) | (t - "A" + 10)
      "a".."f": hex := (hex << 4) | (t - "a" + 10)
      other: abort t

PUB sendChar(c)                                     '' Send a character to the uALFAT
  repeat while ina[uALFAT_Bsy]                      ' Wait for uALFAT not Busy
  repeat while long[ldr#ioControl] & ldr#ioTestRdy  ' Wait for previous I/O to finish
  long[ldr#ioControl+4] := (1 << 16) | @c
  long[ldr#ioControl] := (ldr#ioWrite0Cmd << 24) | uALFAT_Addr
  repeat while long[ldr#ioControl] & ldr#ioTestRdy  ' Wait for this to finish
  return (long[ldr#ioControl] & ldr#ioTestErr) == 0 ' Return true if successful

PUB sendStr(s) | c                                  '' Send a string to the uALFAT
  result := true                                    ' Return true if successful
  repeat while (c := byte[s++])
    if not (result := sendChar(c))
      quit

PUB sendHex(v, w) | s, t                            '' Send a fixed width hex value
  result := true                                    ' Return true if successful
  repeat s from (w - 1) * 4 to 0 step 4
    if (t := ((v >> s) & $F)) < 10
      result := sendChar(t + "0")                   ' "0".."9"
    else
      result := sendChar(t - 10 + "A")              ' "A".."F"
    if not result
      quit
  
PUB recvChar(ms)                                    '' Receive a character from the uALFAT
  if ina[uALFAT_Rdy] == 0                           ' See if character is present
    return -1
  repeat while long[ldr#ioControl] & ldr#ioTestRdy  ' Wait for previous I/O to finish
  long[ldr#ioControl+4] := (1 << 16) | @result
  long[ldr#ioControl] := (ldr#ioRead0Cmd << 24) | uALFAT_Addr | 1
  repeat while long[ldr#ioControl] & ldr#ioTestRdy  ' Wait for this to finish
  if (long[ldr#ioControl] & ldr#ioTestErr) <> 0     ' Return true if unsuccessful
    return -1